import datetime
import json
import logging
import operator
from functools import partial

import yaml
import django
from django.contrib.contenttypes.fields import GenericForeignKey

if django.VERSION[0] >= 3:
    from django.db.models import JSONField
else:
    from django.contrib.postgres.fields.jsonb import JSONField
from django.core.exceptions import ValidationError, FieldDoesNotExist
from django.core.management.color import no_style
from django.core.serializers import python as serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.db import connection
from django.db.models import ForeignKey, TextField, CharField
from django.db.models import fields as django_fields
from django.utils.functional import keep_lazy_text
from tabulate import tabulate

NoneType = type(None)

logger = logging.getLogger(__name__)


def get_field(model, field):
    """
    Recursive model field finder, you can pass it something simple like this:
      > get_field(Model1, 'field1')
      <DjangoField: field1>
    Or complex like this:
      > get_field(Model2, 'model1__model3__model4__id')
      <DjangoField: model4>
    """
    field, _, related = field.partition('__')
    Field = model._meta.get_field(field)
    if not related:
        return Field
    return get_field(Field.related_model, related)


def extract_field_value(data_or_model, field, raise_exceptions=False):
    """
    Similar to get_field() but this works on the gherkin provided data
    """
    if not raise_exceptions and data_or_model is None:
        return data_or_model

    get_val = getattr
    if isinstance(data_or_model, dict):
        get_val = operator.getitem
        if field in data_or_model:
            return data_or_model[field]

    try:
        if '__' not in field:
            val = get_val(data_or_model, field)
            return str(val) if isinstance(val, (dict, list)) and not isinstance(data_or_model._meta.get_field(field), JSONField) else val
        base_field, _, remaining_field = field.partition('__')
        return extract_field_value(get_val(data_or_model, base_field), remaining_field, raise_exceptions)
    except AttributeError:
        logger.exception('')
        if raise_exceptions:
            raise
        return None


@keep_lazy_text
def pretty_print_table(fields, data):
    """
    Stitching two lists, fields + data; the reason is that if we give the fields/headers to tabulate, it will add a
    weird line under the fields/headers that gherkin doesn't like and would complicate the code in order to remove it.
    """
    try:
        rows = [fields] + [[extract_field_value(row, key) for key in fields] for row in data]
    except AttributeError:
        # We could't format it, just dump it.
        return f'\n\nWhat we actually got back was:\n{json.dumps(data, indent=2, cls=DjangoJSONEncoder)}'

    pretty_table = tabulate(rows, tablefmt='orgtbl', numalign='left')
    return f'\n\nWhat we actually got back was:\n{pretty_table}'


class GenericSerializer(serializers.Serializer):
    """
    Serializer to convert the GenericForeignKey fields to natural keys
    """
    def serialize(self, *args, **kwargs):
        self.selected_all_fields = kwargs['fields']
        kwargs['fields'] = [field for field in kwargs['fields'] if field not in '__']
        return super().serialize(*args, **kwargs)

    def handle_fk_field(self, obj, field):
        """
        We convert any nk value entered as `[NK]` into what natural key methods expect: `['NK']`.

        We also convert values for ContentType fields, where the natural key consists of more than one field. We allow
        the developer to enter `[poll, Question]` as the value for the content type field in the test. We then convert
        that value to what content type really expects: `['poll', 'Question']`.
        """
        super().handle_fk_field(obj, field)

        # When serializing models from our testing MigrationGherkin framework, those models will never have
        # natural_key() or get_by_natural_key() in the model's Manager because the model is built dynamically from
        # the current migration script and all prior migration scripts. This extra step compensates for that.
        if obj.__module__ == '__fake__':
            related = getattr(obj, field.name)
            model_name = related.__class__.__name__
            if related:
                if hasattr(related, 'nk'):
                    self._current[field.name] = [related.nk]
                elif model_name == 'ContentType':
                    self._current[field.name] = [related.app_label, related.model]
                else:
                    raise AttributeError(f'Fake model {model_name} does not have a natural key')

        if self._current[field.name]:
            self._current[field.name] = yaml.dump(list(self._current[field.name])).strip()

    def start_object(self, obj):
        self._current = {}

    def get_dump_object(self, obj):
        missing_fields = set(self._current.keys()) ^ set(self.selected_fields)
        for field_name in missing_fields:
            field = get_field(obj, field_name)
            if isinstance(field, GenericForeignKey):
                data = [yaml.dump(list(extract_field_value(obj, field.ct_field).natural_key())).strip(),
                        yaml.dump(list(extract_field_value(obj, field.name).natural_key())).strip()]
                self._current[field.name] = f'[{", ".join(data)}]'
            else:
                self._current[field_name] = extract_field_value(obj, field_name)
        return self._current


class ParseQuery:
    """
    Serializes the queryset and parses the fields into pretty dicts to be used by the tabulate function

    Note that the name is this way because this class is more of a callable than an instance.
    See https://www.python.org/dev/peps/pep-0008/#class-names for more details.
    """
    serialize = partial(GenericSerializer().serialize, use_natural_primary_keys=True, use_natural_foreign_keys=True)

    def __init__(self, queryset, *fields):
        self.queryset = queryset
        self.fields = fields

    def __iter__(self):
        for data in self.serialize(self.queryset, fields=self.fields):
            yield data


def parse_step_objects(context, Model, raise_exceptions=True):
    """
    Parses a table that represents a model/factory.

    This can be used for both display or create/update of models/factories.

    :param Model Model: Model the aloe step table represents
    :param bool raise_exceptions: When doing NK lookups should DoesNotExist be thrown
    :return: yields step row dict with python data
    """
    obj_by_nk_by_field = {}  # {field_name: {nk_value: fk_object}}
    obj_by_id_by_field = {}  # {field_name: {id_value: fk_object}}
    fk_model_by_field = {}  # {field_name: fk_model}
    generic_foreign_key_fields = set()

    # Examine each field specified in the hash table.
    # If it is a generic foreign key, add it to generic_foreign_key_fields.
    # If it is an FK supporting natural keys, add it to obj_by_nk_by_field.
    for field_name in context.table.headings:
        field = get_field(Model, field_name)
        if isinstance(field, GenericForeignKey):
            # GenericForeignKey needs special treatment since there are two pieces to it: 1) the content_type_field,
            # and 2) the object_id_field. Together they evaluate to a single object and that object is the one we care
            # about.
            ForeignModel = Model._meta.get_field(field.ct_field).related_model
            generic_foreign_key_fields.add(field_name)
            is_generic_foreign_key = True
        else:
            ForeignModel = field.related_model  # This is null if this field is not a ForeignKey
            is_generic_foreign_key = False

        if ForeignModel:
            # GenericForeignKey cannot be referenced by ids
            if not is_generic_foreign_key:
                # Tags the field as needing to query all ids in order to show/evaluate the table properly
                obj_by_id_by_field[field_name] = {}
                fk_model_by_field[field_name] = ForeignModel

            # Tags the field as having the ability to search for/display NKs
            foreign_model_fields = [x.name for x in ForeignModel._meta.get_fields()]
            if hasattr(ForeignModel.objects, 'get_by_natural_key') or 'nk' in foreign_model_fields:
                obj_by_nk_by_field[field_name] = {}
                fk_model_by_field[field_name] = ForeignModel

    # Populate a lookup dicts with all the nks and ids that need to be filled in with real models
    for _item in context.table.rows:
        item = _item.as_dict()
        for lookup_field in obj_by_nk_by_field.keys():
            value = item[lookup_field]
            if not value:
                # No need to look for something that's not there..
                pass
            elif value.startswith('[') and value.endswith(']'):
                # Handle special "[..]" nk format
                obj_by_nk_by_field[lookup_field][value] = yaml.load(value, Loader=yaml.FullLoader)
            elif not value.isdigit():
                raise ValueError(f'Please specify a natural key as "[{value}]"')
            else:
                # If the number is a digit, it's probably a literal id -- thus we can ignore the lookup
                pass

    for lookup_field in obj_by_id_by_field.keys():
            value = item[lookup_field]
            if value.isdigit():
                obj_by_id_by_field[lookup_field][value] = None

    # Do a single lookup call for all ids at once.
    for lookup_field, value_map in obj_by_id_by_field.items():
        for instance in get_field(Model, lookup_field).related_model.objects.filter(pk__in=value_map.keys()):
            value_map[str(instance.id)] = instance

    # Find the FK object for each nk.
    for field_name, obj_by_nk in obj_by_nk_by_field.items():
        queryset = fk_model_by_field[field_name].objects
        for original, nk_args in list(obj_by_nk.items()):
            if field_name in generic_foreign_key_fields:
                # Process a generic foreign key field.
                try:
                    # The content type is the first element in the list of nk args for a generic foreign key.
                    # The first element has a nested list of args: app, model_name
                    content_type = queryset.get_by_natural_key(*nk_args[0])

                    # The nk value(s) is the second element in the list of nk args for a generic foreign key.
                    # If it's a simple string, the nk has one element. Otherwise it's a list of nk elements.
                    nk_elements = nk_args[1]
                    if isinstance(nk_elements, str):
                        nk_elements = [nk_elements]

                    # Get the object the generic foreign key points to.
                    obj = content_type.model_class().objects.get_by_natural_key(*nk_elements)
                    obj_id = obj.pk
                except fk_model_by_field[field_name].DoesNotExist:
                    if raise_exceptions:
                        raise
                    content_type = obj = obj_id = None

                # The generic foreign key field actually points to 2 other fields that are persisted to the DB:
                # ct_field: holds the id of the content type of the Model class
                # fk_field: holds the pk of the actual object_ in that Model class
                model_field = get_field(Model, field_name)
                obj_by_nk_by_field[field_name][original] = ReplaceField({
                    model_field.ct_field: content_type,
                    model_field.fk_field: obj_id,
                    field_name: obj
                })
            else:
                try:
                    value = queryset.get_by_natural_key(*nk_args)
                except fk_model_by_field[field_name].DoesNotExist:
                    value = None
                    if raise_exceptions:
                        raise fk_model_by_field[field_name]\
                            .DoesNotExist(f'{fk_model_by_field[field_name]} matching {nk_args} does not exist.')
                except AttributeError:
                    value = queryset.get(nk=nk_args[0])
                # Process a simple foreign key field.
                obj_by_nk_by_field[field_name][original] = value

    for row in context.table.rows:
        value_by_field = {}
        # Convert the string value specified for each field in the hash table to its true python value object.
        for field in row.headings:
            str_value = row.get(field)
            if str_value in obj_by_nk_by_field.get(field, {}):
                # If we found an object by its natural key or generic foreign key, use that object.
                # If no object was found by natural key or generic foreign key, use the key as the value so when we
                # get an error, the error will show the key value.
                value = obj_by_nk_by_field[field].get(str_value, str_value)
            elif str_value in obj_by_id_by_field.get(field, {}):
                value = obj_by_id_by_field[field].get(str_value, str_value)
                # Need to rename the field, there is no concept of '<field>_id' only '<field>'
                field = get_field(Model, field).name
            else:
                # If the field was not a natural key or generic foreign key, the field converts it to its python value.
                value = get_to_python_field(Model, field, str_value)

            if isinstance(value, ReplaceField):
                # If the value is a ReplaceField object, it actually holds the value for several fields.
                # Populate each field the ReplaceField object holds.
                value_by_field.update(value)
            else:
                # Otherwise we have a single value for a single field.
                value_by_field[field] = value
        yield value_by_field


class ReplaceField(dict):
    """
    Special dictionary object used to tag a field that needs to be ignored completely and replaced with whatever this
    holds in it's content. This is mainly used to replace fake/private fields like "GenericForeignKey".
    """
    pass


def get_to_python_field(model, field, value):
    """
    Model.Meta.field.to_python() override, there are some to_python functions that don't behave as expected, this
    function is here to ensure that the correct outcome is returned.

    Example: ForeignKey.to_python() returns the exact value you send it, if you send it a dict, it sends back the same
             dict, nothing changed. This is a problem when you have a number coalesced as a string and you need it to
             be converted to a integer.
    """
    field = get_field(model, field)
    if isinstance(field, ForeignKey):
        # ForeignKey don't like empty strings, they need to be converted to None
        if value == '':
            return None
        field = field.related_model._meta.pk
    elif isinstance(field, (TextField, CharField)):
        if value == '':
            return None
        # Since an empty value is always converted to None, an empty string must be specified as '""' or "''".
        elif value == '""' or value == "''":
            return ''
    elif isinstance(field, JSONField):
        # JSONField.to_python() is not implemented. See the reason at https://code.djangoproject.com/ticket/29147
        # Handle this case to fulfill our own needs.
        return json.loads(value)
    try:
        return field.to_python(value)
    except ValidationError:
        return None


def reset_db_seq(Model, next_value=None):
    """
    Reset a DB table's sequence to desired next_value.

    :param Model :type Model
    :param next_value :type int. If specified, set the sequence so the next value assigned is next_value.
                                 If unspecified, set the sequence so the next value is max(id) + 1.
    """
    try:
        Model._meta.get_field('id')
    except FieldDoesNotExist:
        return  # If id field does not exist, do not reset the sequence.

    with connection.cursor() as cursor:
        if not next_value:
            # Find the highest id value in use if next_value is None or 0.
            cursor.execute(f"SELECT MAX(id) FROM {Model._meta.db_table}")
            highest_value = cursor.fetchone()[0]
            if not isinstance(highest_value, int):
                return

            # If the table has a max(id) of None, it is empty so start the next value at 1.
            # If the table has a max(id) of 1, start the next value at 2.
            # If the table has a max(id) of 3, start the next value at 4.
            highest_value = highest_value or 0
            next_value = highest_value + 1
        # The 3rd parameter we send to setval() specifies whether the specified value has already been used.
        # If false, the next value returned will be the specified value.
        # If true,  the next value returned will be the specified value + 1.
        # https://www.postgresql.org/docs/9.6/static/functions-sequence.html
        # cursor.execute(f"SELECT NVL('{next_value}) {Model._meta.db_table}_id_seq', %s, %s)", (next_value, value_was_used))
        sequence_sql = connection.ops.sequence_reset_sql(no_style(), [Model])
        for sql in sequence_sql:
            cursor.execute(sql)

    return next_value


def convert_type(obj):
    """
    Given the instance of an object, how do we convert a string to that object.

    Limitations: iterables are NOT supported, you'll get strange results, example type `[1,2,3]` will correctly
                 respond with `list()` BUT when you pass in `'[1,2,3]'` your output will be pure strangeness:
                 `['[', '1', ',', '2', ',', '3', ']']` -- for right now, this can we solved when there is a need
                 for it to behave correctly.

    >>> convert_type(datetime.date(2000,1,1))('2000-1-2')
    datetime.date(2000, 1, 2)
    >>> convert_type('hello')(1)
    '1'
    >>> convert_type(None)('')
    >>> convert_type(1.2)(1)
    1.0
    >>> convert_type(1)('3')
    3
    """
    obj_type = type(obj)
    if obj_type == NoneType:
        return lambda x: None if x == '' else x
    elif obj_type == datetime.datetime:
        return partial(django_fields.DateTimeField.to_python, None)
    elif obj_type == datetime.date:
        return partial(django_fields.DateField.to_python, None)
    return obj_type


def get_model(model):
    try:
        app_label, model_name, through_model = model.split('.')
        Model = django.apps.apps.get_model(app_label=app_label, model_name=model_name)
        return getattr(Model, through_model).through
    except ValueError:
        app_label, model_name = model.split('.')
        return django.apps.apps.get_model(app_label=app_label, model_name=model_name)




from behave import *
from django.core import mail
from freezegun import freeze_time


@step('hoy es el "([^"]+)"')
def freeeze(context, date):
    freezer = freeze_time(date, ignore=['behave.runner'])
    freezer.start()


@step('veo que se han enviado emails con los siguientes parametros')
def step_mpl(context):
    context.test.assertEqual(len(mail.outbox), len(context.table.rows))
    for row in context.table.rows:
        mail_send = False
        for email in mail.outbox:
            mail_found = True
            for key, value in row.as_dict().items():
                mail_found &= value == str(getattr(email, key))
            mail_send |= mail_found
        assert mail_send
    # for row, email in zip(context.table.rows, mail.outbox):
    #     for key, value in row.as_dict().items():
    #         assert value == str(getattr(email, key)), f'{key} expected: {value} actual:{getattr(email, key)}'


@step('veo que no se ha enviado email')
def step_mpl(context):
    assert len(mail.outbox) == 0



import inspect
import operator
from ast import literal_eval
from collections import defaultdict
from functools import reduce, partial

import yaml
from behave import *
from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model, models
from django.core.files.base import ContentFile
from django.db import reset_queries, connection
from django.db.models import Q, signals
from django.utils.functional import lazystr
from factory.django import mute_signals
from model_bakery import baker

from .utils import (pretty_print_table, extract_field_value, reset_db_seq, parse_step_objects, ParseQuery,
                    get_model)

UserModel = get_user_model()

all_model_signals = [signal for signal in vars(signals).values() if isinstance(signal, signals.ModelSignal)]


@step('limpio "([^\"]+)"( without resetting its sequence)?')
def clear_the_db(context, model, without_reset_seq):
    Model = get_model(model)
    Model.objects.all().delete()
    if not without_reset_seq:
        reset_db_seq(Model)


@step('reseteo las sequencias de "([^\"]+)"')
def clear_the_db(context, model):
    Model = get_model(model)
    reset_db_seq(Model)


@step('(limpio e )?inserto las siguientes lineas del modelo "([^\"]+)"( sin signals)?')
def insert_to_db(context, limpio, model, no_signals):
    """
    I insert the following rows for "app.Model":
      | nk | name       | age | student__nk |
      | S1 | John Smith | 18  | S1          |
      | S2 | Jane Brown | 19  | S2          |
    I insert the following rows for "app.Model" without signals:
    I clear and insert the following rows for "full.path.to.Factory":
    I clear and insert the following rows for "full.path.to.Factory" without signals:
    """
    Model = get_model(model)

    _all_model_signals = all_model_signals
    if not no_signals:
        # `without signals` means to suppress all signals fired by saving a model row.
        _all_model_signals = []

    if limpio:
        # `clear and` means to delete all rows in the table first.
        # Use the Model to delete all rows.
        Model.objects.all().delete()
        reset_db_seq(Model, next_value=1)

    many_to_many_names = {field.name: field.related_model for field in Model._meta.many_to_many}
    with mute_signals(*_all_model_signals):
        for data in list(parse_step_objects(context, Model, raise_exceptions=False)):
            for field in context.table.headings:
                if field in many_to_many_names.keys():
                    data[field] = many_to_many_names[field].objects.filter(pk__in=literal_eval(data[field]))
            baker.make(Model, **data, _create_files=True)
    reset_db_seq(Model)


@step('modifico las siguientes lineas del modelo "([^\"]+)" identificadas por "([^\"]+)"( sin signals)?')
def update_row_in_db(context, model, filter_fields, without_signals):
    """
    Update existing table rows.

    `without signals` means to suppress all signals fired by saving a model row.

    Examples:
      I update the following rows for "app.Model" identified by "nk, name":
        | nk | name       | age |
        | S1 | John Smith | 18  |
        | S2 | Jane Brown | 19  |
      I update the following rows for "app.Model" identified by "nk" without signals:

    Please note, you CANNOT do multiple field nesting, example: `foreign_field__attribute=False`, you will need to
    update `attribute` in a separate steps.
    """
    Model = get_model(model)
    # The fields that identify each row are separated by a comma.
    filter_fields = [field.strip() for field in filter_fields.split(',')]

    for value_by_field in parse_step_objects(context, Model):
        # Remove the filtering (identifying) field from the data values.
        filters = {key: value_by_field.pop(key) for key in filter_fields}
        # Get the row identified by the filter fields, and update it with the data field values.
        Model.objects.filter(**filters).update(**value_by_field)
        if not without_signals:
            # qs.update does not call signals, but qs.save does.
            Model.objects.get(**filters).save()


@step('hay "([0-9]+)" "([^\"]+)" en base de datos')
def step_impl(context, count, model):
    Model = get_model(model)
    assert Model.objects.count() == int(count)


@step('"([^\"]+)" (tiene exactamente|contiene) las siguientes lineas identificadas por "([^\"]+)"(?: ordenadas por "([^\"]+)")?')
def database_has_rows(context, model, exact_contain, filter_fields, order_fields):
    """
    Verify the contents of a table.

    Examples:
      Then "app.Model" has exactly the following rows identified by "id":
        | id | name       | age | student |
        | 1  | John Smith | 18  | [S1]    |
        | 2  | Jane Brown | 19  | [S2]    |
      Then "app.Model" contains the following rows identified by "id, age":
        | id | name       | age | student |
        | 1  | John Smith | 18  | [S1]    |

      # Note: We can traverse field and check foreign key attributes
      Then "app.Model" contains the following rows identified by "id, age":
        | id | name       | age | student__email   |
        | 1  | John Smith | 18  | jsmith@schoo.com |

    """
    Model = get_model(model)

    fields = context.table.headings
    filter_fields = [field.strip() for field in filter_fields.split(',')]

    actual_values_by_id = {}
    queryset = Model.objects.select_related()
    if order_fields:
        order_fields = [field.strip() for field in order_fields.split(',')]
        queryset = queryset.order_by(*order_fields)
    else:
        queryset = queryset.order_by('pk')
    # Only filter the query if it's a contain, otherwise we'll always retrieve the entire table contents
    if exact_contain == 'contiene':
        filters = []
        for row in context.table.rows:
            _filter = {}
            for field in filter_fields:
                if field in fields:
                    _filter[field] = row[field]
                else:
                    raise KeyError(f'Rows cannot be identified by "{field}". '
                                   f'Ensure that "{field}" is present in the hash table.')

                ForeignModel = getattr(queryset.model._meta.get_field(field), 'related_model')
                if ForeignModel and hasattr(ForeignModel.objects, 'get_by_natural_key'):
                    keys = inspect.getfullargspec(ForeignModel.objects.get_by_natural_key).args[1:]
                    _filter.update(dict(zip([f'{field}__{k}' for k in keys], yaml.load(_filter.pop(field), Loader=yaml.FullLoader))))
            if _filter:
                filters.append(Q(**_filter))
        queryset = queryset.filter(reduce(operator.or_, filters))

    error_msg = lazystr(partial(pretty_print_table, fields, ParseQuery(queryset, *fields)))

    for model in queryset:
        clean_row = {field: extract_field_value(model, field) for field in fields}
        key = tuple([clean_row[field] for field in filter_fields])
        if key in actual_values_by_id:
            raise Model.MultipleObjectsReturned(f'Uniquely identifying rows by {filter_fields} is not enough. '
                                                f'Specify fields whose combination is guaranteed to be unique.')
        actual_values_by_id[key] = clean_row

    # hashes = [{'student': <Student ...>, 'age': 18, 'nk': 'A1'}, {'student': <Student ...>, 'age': 19, 'nk': 'A2'}]
    # Note: all the row values are already converted to Python values
    try:
        hashes = list(parse_step_objects(context, Model, raise_exceptions=False))
    except ValueError:
        print(error_msg)
        raise

    expected_values_by_id = {}
    # Clean up the values that will then be compared, aloe gives it all to us in string format, using the model
    # we then convert it to a python value.
    for row_dict in hashes:
        clean_row = {field: extract_field_value(row_dict, field) for field in fields}
        key = tuple([clean_row[key] for key in filter_fields])
        expected_values_by_id[key] = clean_row
    # Ensure that all the ids we got back are the same as what we expected
    assert actual_values_by_id.keys() == expected_values_by_id.keys(), (actual_values_by_id.keys(), expected_values_by_id.keys())

    # Ensure that all the rows match one by one
    for id_ in expected_values_by_id.keys():
        assert actual_values_by_id[id_] == expected_values_by_id[id_], (actual_values_by_id[id_], expected_values_by_id[id_])


@step('limpio la cache de las queries de base de datos')
def clear_database_query_cache(context):
    reset_queries()


@step('veo que se han hecho "([0-9]+)" queries en base de datos(?: ignorando)?')
def confirm_num_database_queries(context, expected_queries):
    """
    Allows the user to get an accurate count of the amount of queries ran.

    The user can also choose to exclude certain queries that are not wanted in the query count.
    Some examples are:

        # A basic example that ensures that a total of 5 queries ran:
        Then I see that "5" queries were made on the database

        # A more advanced example that ensures that only 7 queries were counted if we IGNORED all queries that
        # started with "SET search_path" OR contained "SAVEPOINT" (these are case insensitive comparisons)
        # There is no limit to the number of rows. Any row that succeeds will eliminate the query.
        # The query column must contain a string
        method that takes one arg.
        # The value column is the value of the arg to pass into the string method.
        Then I see that "7" queries were made on the database while ignoring:
          | method     | value           |
          | startswith | SET search_path |
          | startswith | SET schema      |
          | contains   | savepoint       |
    """
    if getattr(settings, 'SKIP_QUERY_COUNT', False):
        return True
    included_queries = []
    excluded_queries = []
    values_by_method = defaultdict(list)  # key: str method that takes one arg. values: list of args to try 1 at a time
    if getattr(context, 'table'):
        for row in context.table.rows:
            values_by_method[row['metodo']].append(row['valor'].lower())

    for query in connection.queries:
        # Nothing needs to be excluded, just add everything and exit early.
        if not values_by_method:
            included_queries.append(query)
            continue

        include = True
        for method, values in values_by_method.items():
            # If there are any hits, this "query" needs to be ignored
            if hasattr(operator, method):
                # Operators include "contains"
                if any([getattr(operator, method)(query['sql'].lower(), value) for value in values]):
                    include = False
                    break
            else:
                # This gets the specified method name from the str class and evaluates it with one arg: value
                # Example: method='startswith'. 'select * from table'.startswith(value)
                if any([getattr(query['sql'].lower(), method)(value) for value in values]):
                    include = False
                    break

        # Add query to included or excluded list.
        if include:
            included_queries.append(query)
        else:
            excluded_queries.append(query)

    def print_issue():
        pretty_included_queries = '\n'.join([f'{q["sql"]} ({q["time"]})' for q in included_queries])
        pretty_excluded_queries = '\n'.join([f'{q["sql"]} ({q["time"]})' for q in excluded_queries])
        return f'\n{len(included_queries)}\n'\
               f'\n{pretty_included_queries}\n\n' \
               f'------ Excluded queries not in the count ----\n\n{pretty_excluded_queries}'

    assert int(expected_queries) == len(included_queries), print_issue()


class adjust_searchpath_for_model:
    """
    There are some instances when dealing with public tables where the search path must contain only the public schema
    so that we don't unintentionally hit the table with the same name in the tenant schema. ContentType is an example of
    this as it table exists in both the public and tenant schemas. The tables can have very different Django ids for the
    same natural key, so we must use the correct table in the public schema when the model with an FK to ContentType is
    in the public schema. If we don't remove the tenant schema from the search path, the ContentType table in the tenant
    schema will be used as the tenant schema always comes before the public schema when the search path is set.

    Also when manipulating rows in tenant.Member, django-tenant-schemas REQUIRES that only the public schema be on the
    connection's search path. Both the save() and delete() methods will raise an exception if the search path contains
    more than just the public schema. Remove the current tenant schema when entering this context and put it back when
    leaving.
    """

    def __init__(self, Model):
        full_app_path = apps.get_app_config(Model._meta.app_label).name
        self.is_public_model = full_app_path in settings.INSTALLED_APPS

    def __enter__(self):
        if self.is_public_model:
            self.context = schema_context(get_public_schema_name())
            self.context.__enter__()

    def __exit__(self, type, value, traceback):
        if self.is_public_model:
            self.context.__exit__(type, value, traceback)


@step('(limpio y )?asigno los siguientes permisos al usuario con username "([^"]+)"')
def insert_to_db(context, limpio, username):
    user = UserModel.objects.get(username=username)
    if limpio:
        user.user_permissions.all().delete()

    for row in context.table:
        filters = dict(zip(context.table.headings, row))
        user.user_permissions.add(models.Permission.objects.get(**filters))


@step('(limpio y )?asigno los siguientes grupos al usuario con username "([^"]+)"')
def insert_to_db(context, limpio, username):
    user = UserModel.objects.get(username=username)
    if limpio:
        user.groups.all().delete()

    for row in context.table:
        filters = row.as_dict()
        user.groups.add(models.Group.objects.get(**filters))


@step('inserto un file con nombre "([^\"]+)" en el campo "([^\"]+)" del modelo "([^\"]+)" identificado por "([^\"]+)"')
def insert_to_db(context, filename, field, model, filter_fields):
    """
    Y inserto un file con nombre "test.py" en el campo "adjunto" del modelo "permiso.PeticionPermiso" identificado por "id=1"
    """
    Model = get_model(model)
    # The fields that identify each row are separated by a comma.
    filter_fields = dict(x.split("=") for x in filter_fields.split(","))
    instance = Model.objects.get(**filter_fields)
    getattr(instance, field).save(filename, ContentFile(b'asdf'))



"""
Read the docstring in class MockDict for why this file exists.
"""

# Standard library imports.
import inspect
from collections import defaultdict
from copy import deepcopy
from functools import partial, wraps
from os import environ
from types import GeneratorType

from behave import *

from .utils import convert_type


class MockDict:
    """
    Mock registry, keeps track of all the active mocks, either at the scenario level or at the feature/example level.

    The scope of a patch/mock can be for a single scenario, a single feature, or global. Unfortunately any test that
    calls patch.stopall() wipes them all out. No test has the knowledge to know that it is ok to stop absolutely
    everything else. So we must NEVER call patch.stopall().

    This means that every patch/mock must declare/register its desired scope (duration). This allows us to have exactly
    one test step that runs after every scenario to deactivate only the patches/mocks for the current scenario.
    Similarly we can have exactly one test step that runs after every feature to deactivate only the patches/mocks for
    the current feature. This allows us to honor the desired scope of each patch/mock.

    The examples below show how to do this. Note that each mock:
    1) Gives itself a unique name.
    2) Starts its mock(s).
    3) Calls the yield operation to allow the test to continue after starting its mock(s).
    4) Stops its mock(s) after the yield.
       Note that the framework will call the function when it knows the registered scope has ended. The function
       resumes at the statement after the yield, which causes the stop mock logic to run.
    5) The feature file itself specifies the scope of the mock when it invokes the mock.

    Example:

      @mocker('app_label.ThingBeingMockedHumanName')
      def unique_function_name(step_):
          mock = patch('path.to.function.or.Class')
          mock.start()

          yield

          mock.stop()

      # Inside your feature file do:
      I turn on the mock named "app_label.ThingBeingMockedHumanName" in this "scenario"


      @mocker('app.Foo')
      def unique_function_name2(step_, func1_value, func2_value):
          mock1 = patch('path.to.function1', return_value=func1_value)
          mock2 = patch('path.to.function1', return_value=func2_value)
          mock1.start()
          mock2.start()

          yield

          mock1.stop()
          mock2.stop()

      # Inside your feature file do:
      I turn on the mock named "app.Foo" in this "scenario" with arguments:
        | func1_value | func2_value |
        | op1         | op2         |


      @mocker_step(r'I start mock my custom request mock, returning status code "([^\"]+)", and body:$')
      def mock_example(step_, value):
          mock_response = Mock()
          mock_response.json = step_.multiline
          mock_response.status = value
          mock = patch('path.to.function', mock_response)
          mock.start()
          yield
          mock.stop()

       I start mock my custom request mock, returning status code "200", and body:
        '''
        {custom json}
        '''
    """

    def __init__(self):
        self.mocks = {}
        self.active_mock_by_name_by_scope = defaultdict(dict)

    def register(self, name, func):
        """Every mock must be registered with a unique name."""
        if name in self.mocks:
            raise ValueError(f'Mock "{name}" already registered, mock names must be unique')
        self.mocks[name] = func

    def stop(self, scope):
        """
        Call the same function that started the mock. That function has a yield statement, so calling it a second time
        will run the logic after the yield statement that stops the mock.
        TODO Buzzi: Explain why this "stop()" function is different from the "stop_mock()" function. Very confusing.
        """
        active_mock_by_name = self.active_mock_by_name_by_scope[scope]
        for mock_func in active_mock_by_name.values():
            try:
                mock_func()
            except StopIteration:
                # Because the logic after the yield runs without hitting another yield statement, StopIteration will be
                # raised. We expect that.
                pass

        # Now that we've stopped all mocks for this scope, remove them from this scope to show they are not active.
        active_mock_by_name.clear()

    def activate(self, scope, name, func):
        """Start a mock as long as it is not already active. You cannot start a mock twice within the same scope."""
        active_mocks = (mock for scope_ in self.active_mock_by_name_by_scope.values() for mock in scope_.keys())
        # Check that the mock told to run is not currently active
        active_current_mock = (mock for mock in active_mocks if mock == name)

        if any(active_current_mock):
            raise ValueError(f'Mock "{name}" is already active')
        self.active_mock_by_name_by_scope[scope][name] = func

    def stop_mock(self, name):
        # TODO Buzzi: Explain why this "stop()" function is different from the "stop_mock()" function. Very confusing.
        stopped = False
        for scope, active_mock_by_name in self.active_mock_by_name_by_scope.items():
            if name in active_mock_by_name:
                stopped = True
                try:
                    mock_func = active_mock_by_name.pop(name)
                    mock_func()
                except StopIteration:
                    pass
                break
        if not stopped:
            raise ValueError(f'Mock named {name} was never started, cannot be turned off')

    def run(self, scope, name, *args, **kwargs):
        """Check that a mock is registered before activating it."""
        name_ = name.format(*args, **kwargs)

        if name not in self.mocks:
            raise ValueError('Unknown mock. You must register the mock before you use it.')

        self.activate(scope, name_, self.mocks[name](*args, **kwargs))

    def mock(self, name):
        """
        Main wrapper around the functions mocked.

        Note: this is the function that becomes "mocker" decorator
        """

        def wrapper(func):
            @wraps(func)
            def wrapped(*args, **kwargs):
                """
                The wrapped function MUST return a generator, this generator will be called once to start the mock and
                called once more to stop the mocks. It is your responsibility to clean up after your functions.
                """
                # The original generator from the function, note that at this point nothing has been started
                gen = func(*args, **kwargs)
                # We ensure that what we received is indeed a generator, otherwise this won't work
                assert isinstance(gen, GeneratorType), f'Mocked named {name} MUST return a generator'
                # This is the first time we call the function, runs everything above the yield and stops.
                next(gen)
                # We then save the last part of the yield, that will have the clean up code to be run later by either
                # the stop_scenario() or the stop_feature()
                return partial(next, gen)

            # The name is will now be registered, the name MUST be unique!
            self.register(name, wrapped)

            wrapped.original = func

            return wrapped
        return wrapper

    def step(self, step_func_or_sentence, mock_name=None, scope='scenario'):
        """
        This is a wrapper around aloe.step to add our custom code.
        """
        def wrapper(func):
            name = mock_name or f'{func.__module__}.{func.__name__}'
            # Wrap the function around our normal mock() function, to handle the generator logic
            wrapped = self.mock(name)(func)

            @wraps(wrapped)
            def internal(*args, **kwargs):
                self.run(scope, name, *args, **kwargs)

            # Wrap our mock() around func, then wrap that around our mock state management, finally around the
            # standard aloe step(), this is so that we can call it like any other aloe step -- hides the complexity.
            return step(step_func_or_sentence)(internal)

        return wrapper


MOCK_REGISTRY = MockDict()

mocker = MOCK_REGISTRY.mock
mocker_step = MOCK_REGISTRY.step


@step('enciendo el mock llamado "([^\"]+)" para este "(caracteristica|escenario)"( con los argumentos)?')
def start_mocking(context, function_name, scope, receiving_args):
    """
    Starts the named mock functions.

    Examples:
      And I turn on the mock named "app.Foo" in this "feature"

      And I turn on the mock named "app.Foo" in this "scenario" with arguments:
        | named_arg_to_be_passed |
        | example                |

    """
    kwargs = {}
    if receiving_args:
        if context.table:
            assert 1 == len(context.table.rows), 'Mock objects only accept a single of arguments'
            kwargs = context.table.rows[0].as_dict()
    if scope == 'caracteristica':
        MOCK_REGISTRY.run('caracteristica', function_name, context, **kwargs)
    elif scope == 'escenario':
        MOCK_REGISTRY.run('escenario', function_name, context, **kwargs)
    else:
        raise ValueError(f'{scope} is not a valid scope')


@step('apago el mock llamado "([^\"]+)"')
def stop_mocking(step_, name):
    """
    Stops the named mock function given a name.

    If you don't do this, the mock will live for the remainder of it's scope, either feature wide or scenario wide.

    Example:

      And I turn off the mock named "app.Foo"'
    """
    MOCK_REGISTRY.stop_mock(name)


@fixture
def stop_feature_mocks(*args, **kwargs):
    """
    Turns off all the pending feature level mocks.
    """
    MOCK_REGISTRY.stop(scope='caracteristica')


@fixture
def stop_scenario_mocks(*args, **kwargs):
    """
    Turns off all the pending scenario level mocks.

    This will fire after each scenario and will fire after each example inside a scenario.
    """
    MOCK_REGISTRY.stop(scope='escenario')


@step('verifico que el mock "([^\"]+)" ha sido llamado( con los siguientes parametros)?')
def verify_mocked_call_args(context, mock_name, params):
    """
    This step function can be used to assert the parameter with which a mock was called.

    Example:
      @mocker('app_label.ThingBeingMockedHumanName')
      def unique_function_name(step_):
          world.mock_name = mock = patch('path.to.function.or.Class')
          mock.start()

          yield

          mock.stop()

      # Inside your feature file do to activate the mock.
      I turn on the mock named "app_label.ThingBeingMockedHumanName" in this "scenario"

      # To assert mocked function arguments.
      I verify that mock "mock_name" was called with the following parameters:
        | param1 | param2       |
        | 10     | {'test': []} |
        | 2      | {}           |

    """
    patch_obj = getattr(context, mock_name)
    mock_obj = getattr(patch_obj.target, patch_obj.attribute)

    if not params:
        mock_obj.assert_called_once()
        return

    context.test.assertEqual(mock_obj.call_count, len(context.table.rows))

    # Get the original function being patched.
    func = patch_obj.temp_original
    func_signature = inspect.signature(func)
    headers = context.table.headings

    # Return a generator of dict having actual argument name and value pair.
    # def x(a, b, d=1):
    #    pass
    # If x is patched and called as: x(1, 2, 5) then actual_call_args => {a: 1, b:2, d=5}
    def actual_call_args():
        for call in mock_obj.call_args_list:
            call_args = inspect.getcallargs(func, *call[0], **call[1])
            x = {}
            for header in headers:
                if header in call_args.keys():
                    x[header] = call_args[header]
            yield x

    # Generator of dict having expected argument name and value pair.
    expected_call_args = (row.as_dict() for row in context.table.rows)

    for actual, expected in zip(actual_call_args(), expected_call_args):
        for header in headers:
            # Convert expected value from string to whatever type is actual value and assert that they are equal.
            assert actual[header] == convert_type(actual[header])(expected[header]), (actual[header], convert_type(actual[header])(expected[header]))


@mocker_step('(limpio e |)mockeo las siguientes variables de entorno')
def mock_environment_variables(context, clear_environ):
    """ Mock in environment variables to use during tests. """
    old_environ = deepcopy(environ)
    if clear_environ:
        # Only the variables we account for should be there.
        environ.clear()
    for row in context.table.rows:
        # Other environment variables
        row_dict = row.as_dict()
        environ[row_dict['key']] = row_dict['value']

    yield

    # Reset the os.environ dictionary
    environ.clear()
    environ.update(old_environ)



import ast
import json

from behave import *
from django.template import Template, Context
from jq import jq

from .utils import extract_field_value, pretty_print_table


@step('hago las peticiones desde la url "([^"]+)"')
def set_url(context, url):
    try:
        context.http_headers['HTTP_REFERER'] = url
    except AttributeError:
        context.http_headers = {'HTTP_REFERER': url}


@step('hago un "([^"]+)" a la url "([^"]+)"(?: con los argumentos "([^"]+)")?(?: (?:y|con) los parametros "([^"]+)")?(?: (?:y|con) body)?')
def step_impl(context, method_name, url, url_args, url_params):
    """
    Hago un "get" a la url "factura"
    Hago un "get" a la url "factura" con los argumentos "id=1"
    Hago un "get" a la url "factura" con los parametros "ordering=nombre,nombre=mario"
    Hago un "get" a la url "factura" con body:
    Hago un "get" a la url "factura" con los argumentos "id=1" y body:
        | name  |
        | mario |
        | luigi |
    Hago un "get" a la url "factura" con los argumentos "id=1" y body:
        | key   | value |
        | mario | rossi |
        | luigi | verdi |
    Hago un "get" a la url "factura" con los argumentos "id=1" y body:
        '''
        { raw json }
        '''
        # Use triple double-quotes
    """
    headers = {'content_type': 'application/json'}
    try:
        headers = dict(headers, **context.http_headers)
    except AttributeError:
        pass

    data = None
    if context.text:
        data = context.text
    elif context.table:
        fields = context.table.headings
        # This is used to send a single dict as the payload
        if len(fields) == 2 and 'key' in fields and 'value' in fields:
            data = {item['key']: item['value'] for item in context.table.headings}
        else:
            data = list(context.table.headings)
    if not url.startswith('/'):
        # Hack!!
        # Could not find an easier way to parse the named url WITH parameters
        url = Template(f"{{% url '{url}' {url_args or ''} %}}").render(Context())
    if url_params:
        url = f'{url}?{"&".join([param.strip().replace(";", ",") for param in url_params.split(", ")])}'
    method = getattr(context.test.client, method_name.lower())
    context.response = method(url, data=data, **headers)


@step('configuro los headers( usando literales|)')
def add_request_headers(context, use_literals):
    # The way django settings is made, it allows for settings and headers to be passed in as one
    context.http_headers = {}
    cast = ast.literal_eval if bool(use_literals) else lambda x: x
    for item in context.table.rows:
        context.http_headers[item['name']] = cast(item['value'])


@step('el codigo de retorno es "([0-9]{3})"')
def step_impl(context, status_code):
    assert context.response.status_code == int(status_code), (context.response.status_code, context.response.content)


@step('hay "([0-9]+)" elementos en la response')
def step_impl(context, count):
    assert len(context.response.json()['results']) == int(count)


@step('(?:utilizando el formato jq "(.*)")? la response es')
def check_request_response(context, jq_format):
    """
    More details about the jq can be found at their docs: https://stedolan.github.io/jq/manual/#Basicfilters

    List within list

        [
          {"nk": "coursenk1", "prereqs_met": ["coursenkX", "coursenkY"], "prereqs_missed": ["courseidW", "courseidZ"]},
          {"nk": "coursenk2", "prereqs_met": ["coursenkA", "coursenkB"], "prereqs_missed": ["courseidC", "courseidD"]},
        ]

        # And using jq format check the expected values for prereqs_met
        And using jq format "[.[].prereqs_met[] | {value: .}]" the response is:
          | value     |
          | coursenkX |
          | coursenkY |
          | coursenkA |
          | coursenkB |

        # And using jq format check the expected values for prereqs_missed
        And using jq format "[.[].prereqs_missed[] | {value: .}]" the response is:
          | value     |
          | courseidW |
          | courseidZ |
          | courseidC |
          | courseidD |

    List within list within dict

        {
          "courses":
            [
              {"nk": "coursenk1", "prereqs_met": ["coursenkX", "coursenkY"],
               "prereqs_missed": ["courseidW", "courseidZ"]},
              {"nk": "coursenk2", "prereqs_met": ["coursenkA", "coursenkB"],
               "prereqs_missed": ["courseidC", "courseidD"]},
            ],
          "units": 16
        }

        # And using jq format check the expected values for prereqs_met
        And using jq format "[.courses[].prereqs_met[] | {value: .}]" the response is:
          | value     |
          | coursenkX |
          | coursenkY |
          | coursenkA |
          | coursenkB |

        # And using jq format check the expected values for prereqs_missed
        And using jq format "[.courses[].prereqs_missed[] | {value: .}]" the response is:
          | value     |
          | courseidW |
          | courseidZ |
          | courseidC |
          | courseidD |

    Dict within dict

        {
          "courses":
            {
              "MATH101": {"nk": "MATH101", "title": "Algebra"},
              "MATH202": {"nk": "MATH202", "title": "Linear Algebra"},
            },
          "hobbies":
            {
              "Swimming": {"nk": "Swimming", "area": "Outdoors"},
              "Ping Pong": {"nk": "Ping Pong", "area": "Indoors"},
            },
        }

        # And using jq format check the expected values for hobby nk & course nk
        And using jq format "[.[][].nk | {value: .}]" the response is:
          | value     |
          | MATH101   |
          | MATH202   |
          | Swimming  |
          | Ping Pong |

        # And using jq format check the expected values for course nk
        And using jq format "[.courses[].nk | {value: .}]" the response is:
          | value     |
          | Swimming  |
          | Ping Pong |

    """
    is_json = context.response.get('Content-Type') == 'application/json'
    if is_json:
        actual = context.response.json()
        if jq_format:
            # Before trying to debug the lines below, why not test it at this website: https://jqplay.org/
            # It's simpler than adding breakpoints and print statements everywhere...
            # Also, the docs are here: https://stedolan.github.io/jq/manual/#Basicfilters
            try:
                actual = jq(jq_format).transform(actual)
            except StopIteration:
                # jq returns this if nothing was returned
                actual = None
    else:
        actual = context.response.content.decode()

    if context.table:
        fields = context.table.headings
        if isinstance(actual, dict):
            actual = [actual]
        cleaned_data = [{field: str(extract_field_value(item, field)) for field in fields} for item in actual]
        expected_data = [{field: item[field] for field in fields} for item in context.table.rows]
        context.test.assertEquals(cleaned_data, expected_data, pretty_print_table(fields, cleaned_data))
    elif context.text is not None:
        if is_json:
            expected = json.loads(context.text)
        else:
            # Remove all leading whitespace and new line chars from multiline before comparing.
            # Remove whitespace between dict key & value.
            lines = context.text.split('\n')
            lines_without_leading_whitespace = [line.lstrip() for line in lines]
            one_str_of_all_lines = ''.join(lines_without_leading_whitespace)
            without_space_after_colon = one_str_of_all_lines.replace(': ', ':')
            expected = without_space_after_colon
            actual = actual.replace(': ', ':')
        context.test.assertEquals.__self__.maxDiff = None
        context.test.assertEquals(actual, expected)
    else:
        raise Exception('Nothing to compare')


@step('la respuesta contiene los siguientes headers')
def check_headers(context):
    for row in context.table.rows:
        assert row['key'] in context.response.headers, (f'key {row["key"]} not found', context.response.headers.keys())
        assert context.response.headers[row['key']] == row['value'], (context.response.headers[row['key']], row['value'])



import os

from explicit_behave import __version__
from setuptools import find_packages, setup

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-explicit-behave',
    version='.'.join(str(x) for x in __version__),
    description="A collection of explicit behave steps for API testing",
    author="Maurizio Pilia, Javier Buzzi",
    author_email="piliamaurizio@gmail.com, buzzi.javier@gmail.com",
    url="https://github.com/MaurizioPilia/django-explicit-behave",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'behave>=1.2.6',
        'Django>=1.11',
        'freezegun>=0.3.11',
        'PyYAML>=3.13',
        'jq>=0.1.6',
        'factory-boy>=2.12.0',
        'model-bakery>=1.3.2',
        'tabulate>=0.8.3',
    ],
    python_requires='>=3.6',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        "Framework :: Django",
    ],
    zip_safe=False,
)
